use crate::{
    model::{Program, ProgramId, Programs, Section, SectionId, Sections},
    section_runner::SecRunnerState,
    section_runner_json::SecRunnerStateJson,
};
use eyre::WrapErr;
use rumqttc::{LastWill, MqttOptions, QoS};
use std::{
    ops::{Deref, DerefMut},
    sync::Arc,
    time::Duration,
};
use tokio::task::JoinHandle;
use tracing::{debug, info, trace, warn};

#[derive(Clone, Debug)]
struct Topics<T>
where
    T: AsRef<str>,
{
    prefix: T,
}

impl<T> Topics<T>
where
    T: AsRef<str>,
{
    fn new(prefix: T) -> Self {
        Self { prefix }
    }

    fn connected(&self) -> String {
        format!("{}/connected", self.prefix.as_ref())
    }

    fn sections(&self) -> String {
        format!("{}/sections", self.prefix.as_ref())
    }

    fn section_data(&self, section_id: SectionId) -> String {
        format!("{}/sections/{}", self.prefix.as_ref(), section_id)
    }

    fn section_state(&self, section_id: SectionId) -> String {
        format!("{}/sections/{}/state", self.prefix.as_ref(), section_id)
    }

    fn programs(&self) -> String {
        format!("{}/programs", self.prefix.as_ref())
    }

    fn program_data(&self, program_id: ProgramId) -> String {
        format!("{}/programs/{}", self.prefix.as_ref(), program_id)
    }

    fn program_running(&self, program_id: ProgramId) -> String {
        format!("{}/programs/{}/running", self.prefix.as_ref(), program_id)
    }

    fn section_runner(&self) -> String {
        format!("{}/section_runner", self.prefix.as_ref())
    }
}

#[derive(Clone, Debug)]
pub struct Options {
    pub broker_host: String,
    pub broker_port: u16,
    pub device_id: String,
    pub client_id: String,
}

async fn event_loop_task(
    interface: MqttInterface,
    mut event_loop: rumqttc::EventLoop,
) -> eyre::Result<()> {
    use rumqttc::{ConnectionError, Event};
    let reconnect_timeout = Duration::from_secs(5);
    event_loop.set_reconnection_delay(reconnect_timeout);
    loop {
        match event_loop.poll().await {
            Ok(Event::Incoming(incoming)) => {
                debug!(incoming = debug(&incoming), "MQTT incoming message");
                #[allow(clippy::single_match)]
                match incoming {
                    rumqttc::Packet::ConnAck(_) => {
                        info!("MQTT connected");
                        {
                            // HACK: this really should just be await
                            // but that can sometimes deadlock if the publish channel is full
                            let mut interface = interface.clone();
                            let fut = async move { interface.publish_connected(true).await };
                            tokio::spawn(fut);
                        }
                        //.await?;
                    }
                    _ => {}
                }
            }
            Ok(Event::Outgoing(outgoing)) => {
                trace!(outgoing = debug(&outgoing), "MQTT outgoing message");
            }
            Err(ConnectionError::Cancel) => {
                debug!("MQTT disconnecting");
                break;
            }
            Err(err) => {
                warn!("MQTT error, reconnecting: {}", err);
            }
        }
    }
    Ok(())
}

#[derive(Clone)]
pub struct MqttInterface {
    client: rumqttc::AsyncClient,
    topics: Topics<Arc<str>>,
}

impl MqttInterface {
    fn new(options: Options) -> (Self, rumqttc::EventLoop) {
        let mqtt_prefix = format!("devices/{}", options.device_id);
        let topics: Topics<Arc<str>> = Topics::new(mqtt_prefix.into());
        let mut mqtt_opts =
            MqttOptions::new(options.client_id, options.broker_host, options.broker_port);

        let last_will = LastWill::new(topics.connected(), "false", QoS::AtLeastOnce, true);
        mqtt_opts.set_last_will(last_will);

        let (client, event_loop) = rumqttc::AsyncClient::new(mqtt_opts, 16);

        (Self { client, topics }, event_loop)
    }

    async fn publish_data<P>(&mut self, topic: String, payload: &P) -> eyre::Result<()>
    where
        P: serde::Serialize,
    {
        let payload_vec =
            serde_json::to_vec(payload).wrap_err("failed to serialize publish payload")?;
        self.client
            .publish(topic, QoS::AtLeastOnce, true, payload_vec)
            .await
            .wrap_err("failed to publish")?;
        Ok(())
    }

    async fn publish_connected(&mut self, connected: bool) -> eyre::Result<()> {
        self.publish_data(self.topics.connected(), &connected)
            .await
            .wrap_err("failed to publish connected topic")
    }

    async fn cancel(&mut self) -> Result<(), rumqttc::ClientError> {
        self.client.cancel().await
    }

    pub async fn publish_sections(&mut self, sections: &Sections) -> eyre::Result<()> {
        let section_ids: Vec<_> = sections.keys().cloned().collect();
        self.publish_data(self.topics.sections(), &section_ids)
            .await
            .wrap_err("failed to publish section ids")?;
        for section in sections.values() {
            self.publish_section(section).await?;
        }
        Ok(())
    }

    pub async fn publish_section(&mut self, section: &Section) -> eyre::Result<()> {
        self.publish_data(self.topics.section_data(section.id), section)
            .await
            .wrap_err("failed to publish section")
    }

    // Section state can be derived from section runner state...
    pub async fn publish_section_state(
        &mut self,
        section_id: SectionId,
        state: bool,
    ) -> eyre::Result<()> {
        self.publish_data(self.topics.section_state(section_id), &state)
            .await
            .wrap_err("failed to publish section state")
    }

    pub async fn publish_programs(&mut self, programs: &Programs) -> eyre::Result<()> {
        let program_ids: Vec<_> = programs.keys().cloned().collect();
        self.publish_data(self.topics.programs(), &program_ids)
            .await
            .wrap_err("failed to publish program ids")?;
        for program in programs.values() {
            self.publish_program(program).await?;
        }
        Ok(())
    }

    pub async fn publish_program(&mut self, program: &Program) -> eyre::Result<()> {
        self.publish_data(self.topics.program_data(program.id), &program)
            .await
            .wrap_err("failed to publish program")
    }

    pub async fn publish_program_running(
        &mut self,
        program_id: ProgramId,
        running: bool,
    ) -> eyre::Result<()> {
        self.publish_data(self.topics.program_running(program_id), &running)
            .await
            .wrap_err("failed to publish program running")
    }

    pub async fn publish_section_runner(&mut self, sr_state: &SecRunnerState) -> eyre::Result<()> {
        let json: SecRunnerStateJson = sr_state.into();
        self.publish_data(self.topics.section_runner(), &json)
            .await
            .wrap_err("failed to publish section runner")
    }
}

pub struct MqttInterfaceTask {
    interface: MqttInterface,
    join_handle: JoinHandle<()>,
}

impl MqttInterfaceTask {
    pub async fn start(options: Options) -> eyre::Result<Self> {
        let (interface, event_loop) = MqttInterface::new(options);

        let join_handle = tokio::spawn({
            let interface = interface.clone();
            async move {
                event_loop_task(interface, event_loop)
                    .await
                    .expect("error in event loop task")
            }
        });

        Ok(Self {
            interface,
            join_handle,
        })
    }

    pub async fn quit(mut self) -> eyre::Result<()> {
        self.interface.cancel().await?;
        self.join_handle.await?;
        Ok(())
    }
}

impl Deref for MqttInterfaceTask {
    type Target = MqttInterface;

    fn deref(&self) -> &Self::Target {
        &self.interface
    }
}

impl DerefMut for MqttInterfaceTask {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.interface
    }
}
